/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

//#define DEBUG_MODE

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/input/joystick.h"

using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"
#include "../../utils/misc/framerate.h"
#include "../../utils/misc/sim_crap.h"

// cb :
#include "gBase.h"
#include "gUtil.h"
#include "gFloatUtil.h"

/********************************* globals **********************************/
  
int const NumberOfSimSteps = 5;
float const FrameRate = 60.f;
const bool SafeStep = true;

float const dt = 1.f / FrameRate;
double now = 0.0;
float SunRadius = 0.0;

// window/camera stuff
char const AppName[] = "cb2";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = FLT_MAX;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 40.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

void DeleteForce(Binary_Force * force);

bool paused = false;

struct TextureSet;
struct Player;
void ControlShip(Player * player,int pad); 
void SpawnExplosion(Vec2 contactPos,TextureSet * set,float scale=1.f,float duration=1.f);
void SpawnPlanet();

bool NewKeyDown(const int vk);
bool IsKeyDown(const int vk);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.01f, .01f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    float DragLinear;
    float DragAngular;
    
    float SunGravity;
    float SunRadius;
    float SunTexScale;
    float SunGrowRate;
    float SunInsideTexScale;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
    
    //int	NumControls;
    //int	CurControl;
    
    int InitialStars;
    int StarsPerFrame;
    float StarSpawnRadius;
    float StarPointSize;
    float StarStreak;
    
    float PoopLifeTime;
    float PoopOffset;
    float PoopMass;
    float ThrustImpulse;
	float ThrustInterval;
	float TurnRate;
    
	float MissileLifeTime;
	float MissileMass;
	float MissileDisableDuration;
	float ShootInterval;
	float ShootOffset;
	float ShootImpulse;
	
	float ShrapnelMass;
	float ShrapnelVelocity;
	float ShrapnelVelocityFraction;
	float ShrapnelOffset;
	
	float PlanetSpawnRadius;
	float PlanetSpawnDist;
	float PlanetSpawnInterval;
	float PlanetVelocity;
	
	float RamScoopRadius;
	float RamScoopAngle;
	float RamScoopVelocity;
	float EnergyRegenRate;
	float EnergyMax;
	float EnergyBoostRamScoop;
	
	float StarFighterDuration;
	float StarFighterAngleRate;
	float StarFighterMissileLifeTime;
	
	float GrappleKS;
	float GrappleKD;
	
	float PlayerQuadSize;
	
	float SpriteThrustScale;
	float SpriteThrustOffsetX;
	float SpriteThrustOffsetY;
	float SpritePoopScale;
	float SpriteRamScoopFrameRate;
	float SpriteRamScoopScale;
	float SpriteHookFrameRate;
	float SpriteHookScale;
	float SpriteEmpScale;
	float SpriteEmpShrapnelScale;
	float SpriteMissileScale;
	float SpriteMissileShrapnelScale;
//	float GrappleRestLen;
};
static global_properties Globals;
static property *GlobalProperties = NULL;
static property_system_editor *GlobalsEditor = NULL;

static bool MovingCamera = false;
bool autozoom = true;
bool debugstuff = false;
double last_gameover_time = -100.0;
bool GameOver = false;
bool GameOverKillCounted = false;
int GameOverKiller = 2;
bool MenuActive = true;

struct TextureSet
{
	vector<GLuint>	data;
	GLuint Get(float f,int numLoops)
	{
		if ( data.empty() )
		{
			return 0;
		}
		else
		{
			float t = fmodf(f * numLoops,1.f);
			int i = (int)(data.size() * t);
			if ( i == data.size() ) i--;
			return data[i];
		}
	}
	GLuint GetRandom()
	{
		if ( data.empty() )
		{
			return 0;
		}
		else
		{
			int i = (int)data.size() * rand() / RAND_MAX;
			if ( i == data.size() ) i--;
			return data[i];
		}
	}
};

TextureSet * LoadTextureSet(const char * baseName);

struct Sprite
{
	Vec2	pos;
	float	angle;
	float	w,h;
	TextureSet * tex;
	double	start;
	int		numLoops;
	float	duration;
};

GLuint tex_hole = 0;
GLuint tex_glow = 0;
GLuint tex_glow_scooping = 0;
GLuint tex_startup = 0;
GLuint tex_black = 0;
GLuint tex_poop = 0;
GLuint tex_Ship1 = 0;
GLuint tex_Ship2 = 0;
GLuint tex_missile1 = 0;
GLuint tex_missile2 = 0;
GLuint tex_shrapnel1 = 0;
GLuint tex_shrapnel2 = 0;
TextureSet * tex_hook = NULL;
TextureSet * tex_scoop = NULL;
TextureSet * tex_lightning = NULL;
TextureSet * tex_explode = NULL;
TextureSet * tex_missileexplode = NULL;
TextureSet * tex_planet = NULL;

//planet_into_black_hole.wav

char * snd_loop_poop = "content/sounds/misc_e_loop.wav";
char * snd_loop_scoop = "content/sounds/shield_3.wav";
char * snd_loop_grapple = "content/sounds/grappling_hook_1.wav";

char * snd_starfighterMissile = ""; //content/sounds/explosion_5.wav";
char * snd_poop = ""; //content/sounds/explosion_5.wav";

char * snd_intro  = "content/sounds/planet_into_black_hole.wav";
char * snd_endtro = "content/sounds/planet_into_black_hole.wav";

char * snd_fire1 = "content/sounds/missile_1.wav";
char * snd_fire2 = "content/sounds/missile_1.wav";
char * snd_scoopPoop = "content/sounds/misc_d.wav";
char * snd_grappleAttach = "content/sounds/misc_k.wav";
char * snd_grappleFire = "content/sounds/misc_g.wav";
char * snd_missileShrapnel = "content/sounds/explosion_3.wav";
char * snd_shipExplosion = "content/sounds/explosion_5.wav";
char * snd_missileExplosion = "content/sounds/explosion_6.wav";

PROP_STRUCT struct Keymap
{
	int thrust;
	int missile2;
	int missile1;
	int ramscoop;
	int grapple;
	int turnLeft;
	int turnRight;
};

Keymap defaultButtons = {
	Joystick::BUTTON_RIGHT_D,
	Joystick::BUTTON_SHOULDER_L1,
	Joystick::BUTTON_SHOULDER_R1,
	Joystick::BUTTON_SHOULDER_L2,
	Joystick::BUTTON_SHOULDER_R2,
	0,0
};

Keymap defaultKeys1 = {
	'W',
	'R','T',
	'5','6',
	'A','D'
};
Keymap defaultKeys2 = {
	VK_UP,
	'K','L',
	'I','O',
	VK_LEFT,VK_RIGHT
};

Keymap buttons[2] = { defaultButtons, defaultButtons };
Keymap keys[2] = { defaultKeys1, defaultKeys2 };

Keymap * keymaps[4] = { &buttons[0], &buttons[1], &keys[0], &keys[1] };

float player0color[3] = {0.7f,0.65f,0};
float player1color[3] = {0,0.8f,0};

float * playercolors[2] = { player0color, player1color };
	
/*
PROP_STRUCT enum cbControlInput
{
	eLeftStickH,
	eLeftStickV,
	eRightStickH,
	eRightStickV,
	eLeftStickAngle,
	eRightStickAngle,
	eButton_X,
	eButton_O,
	eButton_Square,
	eButton_Triangle,
	eMouseH,
	eMouseV,
	eKey,
	eNull
};

PROP_STRUCT enum cbControlType
{
	eSpringLength,			// spring name , low & high values
	eAngle,					// angle-spring name
};

PROP_STRUCT enum cbControlOrder
{
	eControlValue,	// stick value sets controlled value
	eControlDelta	// stick value adds to controlled value
};

PROP_STRUCT struct cbControlOptions
{
	cbControlInput	m_input;
	cbControlType	m_type;
	cbControlOrder	m_order;
	float			m_controlDeltaSpeed;
	float			m_controlDeltaLo;
	float			m_controlDeltaHi;
	float			m_controlLo;
	float			m_controlRest;
	float			m_controlHi;
	std::string		m_name;
	int				m_gamepad;
	std::string		m_key;
//	char *			m_name;
	
	cbControlOptions();
};

cbControlOptions defaultOptions;

cbControlOptions::cbControlOptions() :
		m_input(eLeftStickH),
		m_type(eSpringLength),
		m_order(eControlValue),
		m_controlDeltaSpeed(1.f),
		m_controlDeltaLo(-FLT_MAX),
		m_controlDeltaHi(FLT_MAX),
		m_controlLo(0.f),
		m_controlRest(1.f),
		m_controlHi(10.f),
		m_name("force0000"),
		m_gamepad(0)
{
}
*/

//static property *cbControlProperties;
//property_system_editor * cbControlEditor;

/*
struct cbControl
{
	cbControlOptions	m_options;
	float				m_value;
	
	cbControl() :
		m_options(),
		m_value(0.f)
	{
	}
	
	void Tick();
};
*/

struct Player;

Vec2 get_facing(Rigid_Body * obj)
{
	float angle = obj->O.z() + PI/2;
	Vec2 facing( cosf( angle ), sinf( angle ) );
	return facing;
}

Vec2 get_perp(Vec2 in)
{
	return Vec2(in.y(), - in.x());
}

float GetRadius(Rigid_Body * rb)
{
	Body * b = rb->m_Pbody.get();
	
	Body_Sphere * sphere = dynamic_cast<Body_Sphere *>(b);
	Body_Box * box = dynamic_cast<Body_Box *>(b);
	if ( sphere )
	{
		return sphere->r();
	}
	else if ( box )
	{
		return box->bs_vert(0).len() * 0.70711f;
	}
	else
	{
		return 1.f;
	}
}

bool IsCircle(Rigid_Body * rb)
{
	Body * b = rb->m_Pbody.get();
	
	Body_Sphere * sphere = dynamic_cast<Body_Sphere *>(b);
	if ( sphere )
	{
		return true;
	}
	return false;
}

// game stuff
struct game_object 
{
    game_object( char *texture_file , Rigid_Body *rbInit ) :
            rb(rbInit)
        {
            assert(rb);

			rb->m_pAppData = this;

            simulator.add(*rb);
            
            spawn_time = GetTime();
            death_time = DBL_MAX;
			is_poop = false;
			is_shrapnel = false;
			ram_scooping = false;
			is_scooping_poops = false;
			is_armed_missile = false;
			is_grapple_head = false;
			player = NULL;
			texture = 0;
            
            quad_size = 1.f;
            
            if ( rb )
            {
				// try to figure out quad_size
				quad_size = GetRadius(rb);
            }
            
            color[0] = 0;
            color[1] = 0;
            color[2] = 0;
        }

    ~game_object( void )
        {
			if ( rb )
			{
				rb->m_pAppData = NULL;
				simulator.remove(*rb);
				delete rb;
			}
			rb = NULL;
        }
        
    Rigid_Body * rb;
    Player *	player;
	double		spawn_time;
	double		death_time;
	float		color[3];
	bool		is_poop;
	bool		is_shrapnel;
	bool		ram_scooping;
	bool		is_scooping_poops;
	bool		is_armed_missile;
	bool		is_grapple_head;
	GLuint		texture;
	float		quad_size;
};

struct Player
{
	game_object * ship;
	double		last_thrust_time;
	double		last_shoot_time;
	double		last_disable_time;
	float		energy;
	bool		is_pooping;
	double		started_starfighter_time;	
	
	Vec2		grapple_gun_facing;
	
	// @@ dangerous pointers :
	game_object * shrapnelly;
	game_object * grapple_head;
	Spring_Force * grapple_force;
	
	Player() : ship(NULL),last_thrust_time(-100.0),last_shoot_time(-100.0), energy(0.0), last_disable_time(-100.0),
				started_starfighter_time(-100.0), grapple_head(NULL),grapple_force(NULL),shrapnelly(NULL)
	{
		is_pooping = false;
		grapple_gun_facing = Vec2(1,0);
	}
};

struct Star
{
	Vec2 x;
	Vec2 v;
};

struct StarFlare
{
	Vec2 x;
	double t;
};

Player players[2];
game_object * pSun = NULL;
Vec2 SunPos;
int player0kills=0,player1kills=0;

vector<Star> starfield;
vector<StarFlare> starflares;
static vector<game_object*> objs;
static vector<Binary_Force *> forces;
//static vector<cbControl> controls;
vector<Sprite> sprites;

vector<game_object *> deferred_delete_list;

void MakeControls();
void TickControls();
void SaveControls();
void LoadControls();

bool TickObject(game_object * obj);
bool TickStar(Star & star);
void SpawnStar();
void BlowUpObject(game_object * obj);
void DeleteObject(game_object * go);

void SpawnShrapnel(game_object * from,const int shrapnel_count,float ShrapnelOffset,
							float ShrapnelMass,float ShrapnelVelocityFraction,
							float ShrapnelVelocity,GLuint tex,float Scale);
							
static Vec2 window_mouse, world_mouse;
static Vec2 delta_window_mouse,delta_world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

bool IsGrappledPlayer(game_object * go)
{
	if ( ! go->player )
		return false;

	Player * other = (go->player == &players[0]) ? &players[1] : &players[0];
	if ( ! other->grapple_force )
		return false;

	if ( go->rb == &other->grapple_force->m_b.rb() )
	{
		return true;
	}
	
	return false;
}

bool my_collision_test_callback(const Body& a, const Body& b,void* rb_a, void* rb_b)
{
	Rigid_Body * rba = (Rigid_Body *)rb_a;
	Rigid_Body * rbb = (Rigid_Body *)rb_b;
	game_object * goa = (game_object *)rba->m_pAppData;
	game_object * gob = (game_object *)rbb->m_pAppData;
	if ( goa && gob )
	{
		// ram-scooping player doesn't collide with poop
		if ( (goa->is_poop && gob->ram_scooping) ||
			(gob->is_poop && goa->ram_scooping) )
		{
			return false;
		}
		// grapple head doesn't collide with poops or shrapnels
		if ( ((goa->is_poop || goa->is_shrapnel) && gob->is_grapple_head) ||
			((gob->is_poop || gob->is_shrapnel) && goa->is_grapple_head) )
		{
			return false;
		}
		
		if ( ((goa->is_poop || goa->is_shrapnel) && IsGrappledPlayer(gob)) ||
			((gob->is_poop || gob->is_shrapnel) && IsGrappledPlayer(goa)) )
		{
			return false;
		}
	}

	if ( a.m_collision_plane == -1 || b.m_collision_plane == -1 )
		return true;
	
	//if ( a.m_collision_plane == b.m_collision_plane )
	//	return false;
	
	const u32 special1 = 0xFFFFFFFF ^ (1UL<<30);
	const u32 special2 = 0xFFFFFFFF ^ (1UL<<31);
	
	if ( a.m_collision_plane == special1 && b.m_collision_plane == special1 )
		return false;
	if ( a.m_collision_plane == special2 && b.m_collision_plane == special2 )
		return false;
	
	return ( a.m_collision_plane & b.m_collision_plane ) != 0;
}

//! Send printf-formatted output to the log, using the eGeneric message type.
//  Same as calling Printf(eGeneric, fmt, ...)
void cbLog(const char* fmt, ...)
{
	va_list			ArgPtr;
	char			TempStr[4096];

	va_start(ArgPtr, fmt);
	_vsnprintf(TempStr, 4096, fmt, ArgPtr);
	va_end(ArgPtr);

	OutputDebugString(TempStr);
	printf(TempStr);
}

GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    if ( ! LoadBitmapByExtension(Filename,&result) )
		return 0;

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

TextureSet * LoadTextureSet(const char * baseName)
{
	TextureSet * set = new TextureSet;
	
	int n = 1;
	for(;;)
	{
		char path[1024];
		sprintf(path,"content/bitmaps/%s/%s%04d.psd",baseName,baseName,n);
		GLuint u = LoadTexture(path);
		if ( u == 0 )
			break;
		set->data.push_back(u); 
		n++;
	}
	
	return set;
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        //simulator.remove(*objs[i]->rb);
        delete objs[i];
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }
    
    forces.clear();
    objs.clear();
    starfield.clear();
    sprites.clear();
    deferred_delete_list.clear();
    
    players[0] = Player();
    players[1] = Player();
    
    int nbodies = (int) simulator.m_bodies.size();
    ASSERT( nbodies == 1 );
}

const char * stristr(const char * search_in,const char * search_for)
{
	while ( *search_in )
	{
		if ( toupper(*search_in) == toupper(*search_for) )
		{
			const char * in_ptr = search_in + 1;
			const char * for_ptr= search_for + 1;
			while ( *for_ptr && toupper(*for_ptr) == toupper(*in_ptr) )
			{
				for_ptr++; in_ptr++;
			}
			if ( ! *for_ptr) 
				return search_in;
		}
		search_in++;
	}

	return NULL;
}

game_object * FindObject(const char * name)
{
	for(int i=0;i<objs.size();i++)
	{
		if ( stricmp(objs[i]->rb->name().c_str(),name) == 0 )
		{
			return objs[i];
		}
	}
	
	return NULL;
}

float loader_scale = 1.f/16.f;
float loader_density_scale = 1/16.f;

void LoadTestFile(void)
{
	SunRadius = Globals.SunRadius;

    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(TestLevelFileName, 0,0, loader_scale, loader_density_scale,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
		if ( stristr(rb_list[i]->name().c_str(),"ignore") != NULL )
			continue;		
    
        objs.push_back(new game_object(NULL,
                                       rb_list[i]));

		if ( IsCircle(objs.back()->rb) )
		{
			objs.back()->texture = tex_planet->GetRandom();                                       
		}
		else
		{
			int x = 2;
		}
    }

    for (int i=0; i < num_force; ++i)
    {
		if ( stristr(force_list[i]->name().c_str(),"ignore") != NULL )
			continue;	
			
        forces.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }

	players[0].ship = FindObject( "ship1" );
	ASSERT( players[0].ship != NULL );
	players[0].ship->player = &players[0];
	players[1].ship = FindObject( "ship2" );
	ASSERT( players[1].ship != NULL );
	players[1].ship->player = &players[1];

	players[0].ship->texture = tex_Ship1;
	players[0].ship->quad_size = Globals.PlayerQuadSize;

	players[1].ship->texture = tex_Ship2;
	players[1].ship->quad_size = Globals.PlayerQuadSize;
	
	pSun = FindObject( "sun" );
	pSun->texture = 0;
	
	// make sure the sun is last in the list :
	
	/*
	{
		vector<game_object*>::iterator it = std::find(objs.begin(),objs.end(),pSun);
		ASSERT( it != objs.end() );
		objs.erase(it);
		objs.push_back(pSun);
	
		Body * b = pSun->rb->m_Pbody.get();
		Body_Sphere * sphere = dynamic_cast<Body_Sphere *>(b);
		sphere->set_r( SunRadius * 0.99f );
	}
	*/
	
	SunPos = pSun->rb->x;
	
	pSun->color[0] = 0;
	pSun->color[1] = 0;
	pSun->color[2] = 0;
	
	memcpy(players[0].ship->color,player0color,3*sizeof(float));
	memcpy(players[1].ship->color,player1color,3*sizeof(float));
	
	for(int i=0;i<Globals.InitialStars;i++)
	{
		SpawnStar();
	}
	
	zsoundPlay(snd_intro);

    if ( MenuActive )
    {
		DeleteObject(players[0].ship);
		DeleteObject(players[1].ship);
    }	
}

game_object * MakeMissile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile("missile.physdb", 0,0, loader_scale, loader_density_scale,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

	Rigid_Body * rb = rb_list[0];
	rb->inv_m = 1.f / Globals.MissileMass;

    return new game_object(NULL,rb);
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	PS2Controller.init();
		
    InitializeFrameRate(FrameRate);
  
	g_collision_test_callback = my_collision_test_callback;

	memset(&Globals,0,sizeof(Globals));

    Globals.PoopLifeTime = 1.f;
    Globals.PoopOffset = 1.f;
	Globals.ThrustImpulse = 1.f;
	Globals.TurnRate = 0.f;

    Globals.Gravity = -30.0f;
	Globals.DragLinear = 0.01f;
	Globals.DragAngular = 0.01f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

 //   Globals.NumControls = 1;
 //  Globals.CurControl = 0;
    
    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    
    /*
	cbControlProperties = CloneProperty(PROP_LOCAL_STRUCT(cbControlOptions));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(cbControlOptions), &defaultOptions,
                            cbControlProperties);
    cbControlEditor = OpenPropertySystemEditor("cbControlOptions", 850, 650, 400, 900);
    EditPropertiesInPlace(cbControlEditor, cbControlProperties);
    */
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    LoadControls();
    
    // gl stuff
    glClearColor(0.0f,0.0f,0.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glDisable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "space_song_a.sgt", MUSIC_TRACK_MAIN);
    music.play("background");
                        
    // sound init
    zsoundInit(WindowHandle);

	// prime the sounds :
	//zsoundPlay(snd_loop_grapple,1,0.f);
	//zsoundPlay(snd_loop_scoop,1,0.f);
	//zsoundPlay(snd_loop_grapple,1,0.f);

	zsoundPlay(snd_starfighterMissile,0,0.f);
	zsoundPlay(snd_fire1,0,0.f);
	zsoundPlay(snd_fire2,0,0.f);
	zsoundPlay(snd_scoopPoop,0,0.f);
	zsoundPlay(snd_grappleAttach,0,0.f);
	zsoundPlay(snd_grappleFire,0,0.f);
	zsoundPlay(snd_missileShrapnel,0,0.f);
	zsoundPlay(snd_shipExplosion,0,0.f);
	zsoundPlay(snd_missileExplosion,0,0.f);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);
    
    tex_startup = LoadTexture("content/bitmaps/start.psd");
	tex_hole = LoadTexture("content/bitmaps/blackhole_0001.jpg");
//	tex_hole = LoadTexture("content/bitmaps/blackhole_0001.psd");
	tex_Ship1 = LoadTexture("content/bitmaps/ship_01.psd");
	tex_Ship2 = LoadTexture("content/bitmaps/ship_02.psd");
	tex_black = LoadTexture("content/bitmaps/circle.psd");
	tex_poop = LoadTexture("content/bitmaps/engine_plasma_01.JPG");
	tex_glow = LoadTexture("content/bitmaps/engine.psd");
	tex_glow_scooping = LoadTexture("content/bitmaps/engine_ramscoop.psd");
	
	tex_missile1  = LoadTexture("content/bitmaps/emp.jpg");
	tex_shrapnel1 = LoadTexture("content/bitmaps/emp_shrapnel.jpg");
	tex_missile2  = LoadTexture("content/bitmaps/missile.psd");
	tex_shrapnel2 = LoadTexture("content/bitmaps/missile_shrapnel.psd");

	tex_scoop = LoadTextureSet("ramscoop");
	tex_hook = LoadTextureSet("grapple");
	tex_lightning = LoadTextureSet("lightning");
	tex_explode = LoadTextureSet("explode");
	tex_missileexplode = LoadTextureSet("missileexplode");
	tex_planet = LoadTextureSet("planet");
	
    LoadTestFile();
}

void UpdateControlCounts()
{
/*
	if ( Globals.NumControls < 1 )
	{
		Globals.NumControls = 1;
	}
	if ( Globals.NumControls != controls.size() )
	{
		controls.resize( Globals.NumControls );
	}
	Globals.CurControl = gClamped(Globals.CurControl,0,Globals.NumControls-1);
*/
}

void RefreshEditors()
{
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
	
	if ( GlobalsEditor )
		UpdateDisplay(GlobalsEditor);
	
//    ConvertStructToProperty(PROP_LOCAL_STRUCT(cbControlOptions), &controls[Globals.CurControl],
//                            cbControlProperties);	
//	UpdateDisplay(cbControlEditor);
}

void Reset()
{
    ClearAll();
    LoadTestFile();
        
    /*
    for(int i=0;i<controls.size();i++)
    {
		controls[i].m_value = 0.f;
    }
    */
}
            
//Thing *thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees )

void ProcessContactList()
{
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
		Vec2 contactPos = cit->m_x;
		
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        //if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        ASSERT( r1 && r2 );
        game_object * go1 = (game_object *) r1->m_pAppData;
        game_object * go2 = (game_object *) r2->m_pAppData;
        if ( go1 == NULL || go2 == NULL )
			continue;
        
        if ( r1 == pSun->rb || r2 == pSun->rb )
        {
			// hit the sun !
			game_object * go = (r1 == pSun->rb) ? go2 : go1;

			deferred_delete_list.push_back(go);
			go->rb->m_pAppData = NULL;
			
			continue;
        }

        // look for missile-ship contact (if the missile is still armed)
        if ( (go1->is_armed_missile || go2->is_armed_missile) &&
			(go1->player || go2->player ) )
        {
			game_object * missile = (go1->is_armed_missile) ? go1 : go2;
			ASSERT( missile->is_armed_missile );
			missile->is_armed_missile = false;
			
			game_object * ship = (go1 == missile) ? go2 : go1;
			ASSERT( ship != missile );
			ASSERT( ship->player != NULL );
			
			Player * P = ship->player;
			
			P->last_disable_time = now;
			
			if ( P->grapple_force )
			{
				DeleteForce(P->grapple_force);
			}
			P->grapple_force = NULL;
			P->grapple_head = NULL;
			
			SpawnShrapnel(missile,10,2.f,0.1f,0.5f,10.f,tex_shrapnel1,Globals.SpriteMissileShrapnelScale);
			
			deferred_delete_list.push_back(missile);
			missile->rb->m_pAppData = NULL;
			
			SpawnExplosion(contactPos,tex_missileexplode,1.f,2.f);

			zsoundPlay(snd_missileExplosion);
						
			continue;
        }
        
        
        if ( (go1->is_grapple_head || go2->is_grapple_head) )
        {
			game_object * head = (go1->is_grapple_head) ? go1 : go2;
			ASSERT( head->is_grapple_head );
			
			game_object * target = (head == go1) ? go2 : go1;
			ASSERT( target != head );
			if ( target->is_poop )
				continue;
			
			// connect go1 & go2;
			Player * shooter;
			if ( head == players[0].grapple_head )
				shooter = players+0;
			else if ( head == players[1].grapple_head )
				shooter = players+1;
			else
				continue;
			
			head->is_grapple_head = false;
			
			if ( shooter->grapple_force )
				continue;
			
			float curLen = (shooter->ship->rb->x - contactPos).len();
			
			Spring_Force * force = new Spring_Force( Handle(*shooter->ship->rb, Vec2(0,0)),
                                                Handle(*target->rb, target->rb->world_space_to_body_space(contactPos)),
                                                Globals.GrappleKS, Globals.GrappleKD, curLen );
        
			simulator.add(*force);
			forces.push_back(force);
			
			deferred_delete_list.push_back(head);
			head->rb->m_pAppData = NULL;
			
			shooter->grapple_force = force;
			shooter->grapple_head = NULL;
			
			zsoundPlay(snd_grappleAttach);
		}
        
        /*
        {
			Vec2 contactPointInBodySpace = asker->rb->world_space_to_body_space( cit->m_x );
			float contactAngleRadians = atan2( contactPointInBodySpace.m_v[1], contactPointInBodySpace.m_v[0] );
			float contactAngleDegrees = radiansToDegrees( contactAngleRadians );
			float delta = fabs( angleDegrees - contactAngleDegrees );
			if( delta < angleToleranceDegrees || delta > 360 - angleToleranceDegrees )
			{
				if( asker->rb == r1 )
					return (Thing *)r2->m_pAppData;
				else
					return (Thing *)r1->m_pAppData;
			}
		}
		*/
	}
	//return 0;
}

void DeleteForce(Binary_Force * force)
{
	vector<Binary_Force*>::iterator it = std::find(forces.begin(),forces.end(),force);
	if ( it != forces.end() )
	{
		forces.erase(it);
	}
	simulator.remove(*force);
	delete force;
	
	if ( players[0].grapple_force == force ) players[0].grapple_force = NULL;
	if ( players[1].grapple_force == force ) players[1].grapple_force = NULL;
}

void DeleteObject(game_object * go)
{
	// delete any forces that point at go :
	Rigid_Body * rb = go->rb;
	for(int i=0;i<forces.size();i++)
	{
		if ( &(forces[i]->m_a.rb()) == rb
			|| &(forces[i]->m_b.rb()) == rb )
		{
			DeleteForce(forces[i]);
			i=-1;
		}
	}

	vector<game_object*>::iterator it = std::find(objs.begin(),objs.end(),go);
	if ( it != objs.end() )
	{
		objs.erase(it);
	}
	delete go;
	
	if ( players[0].ship == go ) players[0].ship = NULL;
	if ( players[1].ship == go ) players[1].ship = NULL;
	if ( players[0].grapple_head == go ) players[0].grapple_head = NULL;
	if ( players[1].grapple_head == go ) players[1].grapple_head = NULL;
	if ( players[0].shrapnelly == go ) players[0].shrapnelly = NULL;
	if ( players[1].shrapnelly == go ) players[1].shrapnelly = NULL;
}

void TexturedQuad(const Vec2 pos,float angle,const float w,const float h)
{
	// Ryan's textures face left are sideways
	angle -= PI/2;

	Vec2 n( cosf(angle), sinf(angle) );
	Vec2 p( n.y(), - n.x() );
	
	n *= h;
	p *= w;
	
	Vec2 a = pos + n + p;
	Vec2 b = pos + n - p;
	Vec2 c = pos - n - p;
	Vec2 d = pos - n + p;
	
	glTexCoord2f(1.f,1.f);
	glVertex2fv((float *)&a);
	glTexCoord2f(1.f,0.f);
	glVertex2fv((float *)&b);
	glTexCoord2f(0.f,0.f);
	glVertex2fv((float *)&c);
	glTexCoord2f(0.f,1.f);
	glVertex2fv((float *)&d);
}

void RenderScene();

void DrawStringCentered(float y,char *str)
{
	DrawString(WindowWidth/2 - (float)strlen(str)*10,y,str);
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
	now = GetTime();
	
	if ( ! MenuActive )
	{
		if ( GameOver && ! GameOverKillCounted && (now - last_gameover_time) > 0.5f )
		{
			GameOverKillCounted = true;
			if ( players[0].ship && ! players[1].ship )
			{
				player0kills ++;
				GameOverKiller = 0;
			}
			else 
			if ( players[1].ship && ! players[0].ship )
			{
				player1kills ++;
				GameOverKiller = 1;
			}
			else
			{
				GameOverKiller = 2;
			}
		}
		if ( GameOver && (now - last_gameover_time) > 5 )
		{
			GameOver = false;
			last_gameover_time = -100.0;
			Reset();
		}
	}

	#ifndef DEBUG_MODE
	if ( debugstuff )
		ShowCursor(TRUE);
	else
		ShowCursor(FALSE);
	#endif

	// first copy cur Control to its struct :
	/*
	if ( Globals.CurControl >= 0 && Globals.CurControl < controls.size() )
	{
		ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "cbControlOptions"),
								cbControlProperties,
								&controls[Globals.CurControl]);		
	}
	*/


		
	// update from properties :
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

	//UpdateControlCounts();

	// now look at this control :
	//RefreshEditors();

	gravity.m_g.set_y(Globals.Gravity);
	global_drag = Drag_Force(Globals.DragLinear,Globals.DragAngular);
    
	PS2Controller.sample();


	/**************************** step the simulator ****************************/   

	if ( MenuActive )
	{
		if (PS2Controller.button_went_down(Joystick::BUTTON_START,0) ||
			PS2Controller.button_went_down(Joystick::BUTTON_START,1) ||
			IsKeyDown(' ') || IsKeyDown(VK_RETURN) )
		{
			MenuActive = false;
			Reset();
			GameOver = GameOverKillCounted = false;
			player0kills = player1kills = 0;
		}
	}
	else
	{
		if (PS2Controller.button_went_down(Joystick::BUTTON_START,0) ||
			PS2Controller.button_went_down(Joystick::BUTTON_START,1) )
		{
			paused = ! paused;
		}
		if ( (PS2Controller.button_went_down(Joystick::BUTTON_SELECT,0) ||
			PS2Controller.button_went_down(Joystick::BUTTON_SELECT,1)) && paused )
		{
			Reset();
			paused =false;
		}
	}
	
	if ( ! paused )
	{
		if ( ! MenuActive )
		{
			SunRadius += Globals.SunGrowRate;
		}
			
		// decay poops
		for(int i=(int)objs.size()-1;i>=0;i--)
		{
			if ( now > objs[i]->death_time ||
				TickObject(objs[i]) )
			{
				// die
				DeleteObject(objs[i]);
			}
		}
	
		// spawn planets
		static double last_planet_spawn = -100.0;
		if ( (now - last_planet_spawn) > Globals.PlanetSpawnInterval )
		{
			last_planet_spawn = now;
			SpawnPlanet();
		}
		
		if ( ! MenuActive )
		{
			if ( players[0].ship )
				ControlShip(players+0,0);
			if ( players[1].ship )
				ControlShip(players+1,1);
		}
		
		TickControls();

		if (spring)                             //update
			spring->m_a = Handle(world_body, world_mouse);

		for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) 
		{
			// step as many times as you like
			// @@ safe step causes glitches!
			//simulator.step(1/(NumberOfSimSteps*GetFrameRate()),false);
			simulator.step(1/(NumberOfSimSteps*GetFrameRate()),SafeStep);
			// do any game stuff you want here (like per-step collision response, etc.)
		}
		
		ProcessContactList();
		
		while(!deferred_delete_list.empty())
		{
			game_object * go = deferred_delete_list.back();
			deferred_delete_list.pop_back();
			DeleteObject(go);
		}
	}
	else
	{
		// paused
	}
	
	
	if ( autozoom )
	{
		static Vec2 p0(0,0);
		static Vec2 p1(0,0);
		if ( players[0].ship ) p0 = players[0].ship->rb->x;
		if ( players[1].ship ) p1 = players[1].ship->rb->x;
		
		Vec2 m = (p0 + p1)*0.5f;
		float newCameraX = m.x();
		float newCameraY = m.y();
		
		float d = (p0 - p1).len();
		float newCameraZ = d*0.8f + 20.f;
		newCameraZ = max(newCameraZ,60.f);
	
		float t = 0.4f;
		CameraX = flerp(CameraX,newCameraX,t);
		CameraY = flerp(CameraY,newCameraY,t);
		CameraZ = flerp(CameraZ,newCameraZ,t);
	}
	
   RenderScene();
   
	if ( MenuActive )
	{
		glColor4f(1,0,0,1);
		/*
		DrawStringCentered(WindowHeight/2 - 16 + 40,"Space Game");
		DrawStringCentered(WindowHeight/2 - 16     ,"CB+RE+MS+IGJ2");
		DrawStringCentered(WindowHeight/2 - 16 - 40,"Press Start or Enter");
		*/

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR); // GL_LINEAR_MIPMAP_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	    
		glColor4f(1,1,1,1);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,tex_startup);
		
		PushOrthoPixels();
			
		glBegin(GL_QUADS);
		TexturedQuad( Vec2(WindowWidth*0.5f,WindowHeight*0.5f), PI/2, WindowWidth*0.5f,WindowHeight*0.5f);
		glEnd();		
		
		PopOrthoPixels();
	}
	else
	{
   
		if ( paused )
		{
			glColor4f(1,0,0,1);
			DrawStringCentered((float)WindowHeight/2 - 16,"Paused");
		}
		if ( GameOver && GameOverKillCounted )
		{
			float red[3] = { 1,0,0 };
			float * color = red;
			char * str = "Tie!";
			if ( GameOverKiller == 0 )
			{
				color = player0color;
				str = "Player 0 Kill";
			}
			else if ( GameOverKiller == 1 )
			{
				color = player1color;
				str = "Player 1 Kill";
			}
			
			glColor4f(color[0],color[1],color[2],1);
			DrawStringCentered((float)WindowHeight/2 - 16,str);
		}
	}
	
	// now wait for the rest of the time so we're consistent on fast machines
	WaitFrameRate(GetFrameRate());
		    
    SwapBuffers(WindowDC);
    
}

void RenderScene()
{
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    /*
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();
    */

    // draw mouse
    /*
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();
	*/
	
	//
	
	// draw the star-field :
	
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR); // GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    
	glColor4f(1,1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex_hole);
		
	glBegin(GL_QUADS);
	TexturedQuad(SunPos,0.f,Globals.SunRadius * Globals.SunTexScale,Globals.SunRadius * Globals.SunTexScale);
	glEnd();
	
	float pointSize = Globals.StarPointSize / CameraZ;
    glPointSize(pointSize);
    glLineWidth(pointSize);
    glEnable( GL_POINT_SMOOTH );	
    glEnable(GL_LINE_SMOOTH);
    //glEnable(GL_ANTI_ALIAS)
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);	// Make round points, not square points
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);		// Antialias the lines

	//*
	glDisable(GL_TEXTURE_2D);
    glBegin(GL_LINES);
    glColor3f(1,1,1); 
	
	// tick starfield
	for(int i=(int)starfield.size()-1;i>=0;i--)
	{
		// @@ quads would be better
		Vec2 lx = starfield[i].x - starfield[i].v * Globals.StarStreak;
		glVertex2f(lx.x(),lx.y());
		glVertex2f(starfield[i].x.x(),starfield[i].x.y());
		if ( TickStar(starfield[i]) )
		{
			starflares.push_back( StarFlare() );
			starflares.back().x = starfield[i].x;
			starflares.back().t = now;
			
			starfield[i] = starfield.back();
			starfield.pop_back();
		}
	}
    glEnd();
    /**/
    
	// spawn stars :
	for(int i=0;i<Globals.StarsPerFrame;i++)
	{
		SpawnStar();
	}		
	
	vector<game_object*> edges;
	
	// iterate all the bodies, rendering them
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		float alpha = fclampunit((float)(obj.death_time - now)*2.f/Globals.PoopLifeTime);

		if ( obj.player )
			continue; // skip the player, he'll draw himself

		if ( obj.is_poop )
			continue;

		if ( memcmp(obj.rb->name().c_str(),"edge",4) == 0 )
		{
			edges.push_back(&obj);
			continue;
		}

		if ( obj.is_grapple_head )
		{			
			glEnable(GL_TEXTURE_2D);
			glColor4f(1,1,1,1);
			glBindTexture(GL_TEXTURE_2D,tex_hook->Get((float)now*Globals.SpriteHookFrameRate,1));
			
			glBegin(GL_QUADS);
			TexturedQuad(obj.rb->x,obj.rb->O.z(),obj.quad_size * Globals.SpriteHookScale,obj.quad_size * Globals.SpriteHookScale);
			glEnd();
			
			continue;
		}

		if ( obj.texture == tex_shrapnel1 || obj.texture == tex_missile1 )
		{
			  glBlendFunc(GL_ONE, GL_ONE);
			glEnable(GL_TEXTURE_2D);
			glColor4f(alpha,alpha,alpha,1);
			glBindTexture(GL_TEXTURE_2D,obj.texture);
			
			glBegin(GL_QUADS);
			TexturedQuad(obj.rb->x,obj.rb->O.z(),obj.quad_size,obj.quad_size);
			glEnd();
		   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			continue;
		}
		
		if ( obj.texture )
		{
			glEnable(GL_TEXTURE_2D);
			glColor4f(1,1,1,alpha);
			glBindTexture(GL_TEXTURE_2D,obj.texture);
			
			glBegin(GL_QUADS);
			TexturedQuad(obj.rb->x,obj.rb->O.z(),obj.quad_size,obj.quad_size);
			glEnd();
		}		
		
		if ( ! obj.texture || debugstuff )
		{
			glDisable(GL_TEXTURE_2D);
			glBindTexture(0,0);
			glColor4f(obj.color[0],obj.color[1],obj.color[2],alpha);
		    glBody1v(obj.rb->m_Pbody.get(), true);
		}
    }
    
    // deferred edges
    for(int i = 0;i < edges.size();++i) 
	{
		glDisable(GL_TEXTURE_2D);
		glBindTexture(0,0);
		glColor4f(0,0,0,1);
		glBody1v(edges[i]->rb->m_Pbody.get(), true);
	}
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(1,1,1,1);
	glEnable(GL_TEXTURE_2D);
	for(int p=0;p<2;p++)
	{
		if ( ! players[p].ship)
			continue;
		Player * P = players+p;

		game_object & obj = *(P->ship);
		
		glEnable(GL_TEXTURE_2D);
		glColor4f(1,1,1,1);
		glBindTexture(GL_TEXTURE_2D,obj.texture);
		
		glBegin(GL_QUADS);
		TexturedQuad(obj.rb->x,obj.rb->O.z(),obj.quad_size,obj.quad_size);
		glEnd();
			
		float t = ( (float)(now - P->last_disable_time) / Globals.MissileDisableDuration );
		if ( t >= 0.f && t <= 1.f )
		{
			GLuint tex = tex_lightning->Get(t,3);
					
			float size = obj.quad_size * 1.2f;
			
			glBindTexture(GL_TEXTURE_2D,tex);
			glBegin(GL_QUADS);
			TexturedQuad(obj.rb->x,obj.rb->O.z(),size,size);
			glEnd();
		}
		
		if ( P->ship->ram_scooping )
		{
			float size = obj.quad_size * Globals.SpriteRamScoopScale;
			
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ONE);
			glBindTexture(GL_TEXTURE_2D,tex_scoop->Get((float)now*Globals.SpriteRamScoopFrameRate,1));
			glBegin(GL_QUADS);
			TexturedQuad(obj.rb->x,obj.rb->O.z(),size,size);
			glEnd();	
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);			
		}
		
		if ( P->is_pooping || P->ship->is_scooping_poops )
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ONE);
    
			float size = obj.quad_size * Globals.SpriteThrustScale;
			
			if ( P->ship->is_scooping_poops )
				glBindTexture(GL_TEXTURE_2D,tex_glow_scooping);
			else
				glBindTexture(GL_TEXTURE_2D,tex_glow);

			Vec2 facing = get_facing(obj.rb);
			Vec2 side = get_perp(facing);
			Vec2 pos = obj.rb->x;

			pos += Globals.SpriteThrustOffsetX * side * obj.quad_size;
			pos += Globals.SpriteThrustOffsetY * facing * obj.quad_size;

			glBegin(GL_QUADS);
			TexturedQuad(pos,obj.rb->O.z(),size,size);
			glEnd();		
			
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
		}
	}
	
	// draw sprites :
	glColor4f(1,1,1,1);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR); //GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	for(int i=0;i<sprites.size();i++)
	{
		Sprite & s = sprites[i];
		
		float t = (float)(now - s.start)/s.duration;
		if ( t > 1.f )
		{
			// kill it
			sprites[i] = sprites.back();
			sprites.pop_back();
			i--;
			continue;
		}
		
		GLuint tex = s.tex->Get(t,s.numLoops);
		glBindTexture(GL_TEXTURE_2D,tex);
		
		glBegin(GL_QUADS);
		TexturedQuad(s.pos,s.angle,s.w,s.h);
		glEnd();
	}
	
	// do the poops :
    glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex_poop);
	glBegin(GL_QUADS);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		if ( ! obj.is_poop )
			continue; // only poops

		float alpha = fclampunit((float)(obj.death_time - now)*2.f/Globals.PoopLifeTime);

		glColor4f(obj.color[0]*alpha,obj.color[1]*alpha,obj.color[2]*alpha,1.f);
			
		TexturedQuad(obj.rb->x,obj.rb->O.z(),obj.quad_size * Globals.SpritePoopScale,obj.quad_size * Globals.SpritePoopScale);
	}
	glEnd();
	
	//glColor4f(0,0,0,1);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBindTexture(GL_TEXTURE_2D,tex_black);
		
	glBegin(GL_QUADS);
	TexturedQuad(SunPos,0.f,SunRadius*Globals.SunInsideTexScale,SunRadius*Globals.SunInsideTexScale);
	glEnd();
	
	/*
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_ONE);
	glColor4f(1,1,1,1);
	glBindTexture(GL_TEXTURE_2D,tex_hole);
		
	glBegin(GL_QUADS);
	TexturedQuad(SunPos,0.f,Globals.SunTexRadius,Globals.SunTexRadius);
	glEnd();
	*/
			
	glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBegin(GL_POINTS);
		
	for(int i=(int)starflares.size()-1;i>=0;i--)
	{
		float t = (float)(now - starflares[i].t)/1.f;
		if ( t >= 1 )
		{
			starflares[i] = starflares.back();
			starflares.pop_back();
			continue;
		}
		// @@ do textured quads that grow by t
		glPointSize( 100.f * pointSize * t );
	    glColor4f(1,1,1, (1.f - t) ); 
		glVertex2f(starflares[i].x.x(),starflares[i].x.y());
	}
    
    glEnd();
    
    // 
    
	// draw the grapple line :
	
	glLineWidth(3.f);
	glBegin(GL_LINES);
	for(int p=0;p<2;p++)
	{
		if ( ! players[p].ship)
			continue;
		float * c = players[p].ship->color;
		Player * P = players+p;
		
		Vec2 from = P->ship->rb->x;
		
		Vec2 end;
		if ( P->grapple_force )
		{
			end = P->grapple_force->m_b.x_worldS();
		}
		else if ( P->grapple_head )
		{
			end = P->grapple_head->rb->x;
		}
		else
		{
			continue;
		}
		
		glColor4f(c[0],c[1],c[2],0.5f);
		glVertex2fv((float *)&from);
		glVertex2fv((float *)&end);		
	}
	glEnd();

	// draw the Game HUD (energy bars + kills)
	glDisable(GL_TEXTURE_2D);
   PushOrthoPixels();
	for(int p=0;p<2;p++)
	{
		float * c = (p == 0) ? player0color : player1color;
		glColor4f(c[0],c[1],c[2],0.5f);
		float e = players[p].energy / Globals.EnergyMax;
		
		float y = WindowHeight * e;
		
		float base = (p == 0) ? 10.f : WindowWidth;
		float sign = (p == 0) ? 1.f : -1.f;
		
		glBegin(GL_QUADS);
		glVertex2f(base + sign*10.f,0.f);
		glVertex2f(base + sign*20.f,0.f);
		glVertex2f(base + sign*20.f,y);
		glVertex2f(base + sign*10.f,y);
		glEnd();
	
		int kills = (p == 0) ? player0kills : player1kills;
		char buffer[80];
		sprintf(buffer,"%d",kills);
		float x = base + sign*30.f;
		if ( p == 1 ) x -= 30.f;
		DrawString(x, WindowHeight-50.f, buffer);
	} 
   PopOrthoPixels();
   
    glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	/*
	glColor3f(0, 0, 1);
	glBegin(GL_LINES);
	float StickScale = 10.0f;
	glVertex2f(0, 0);
	glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
	glVertex2f(0, 0);
	glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
	glEnd();
	*/
		
	if ( debugstuff )
	{
		glColor4f(1, 1, 1, 1);
		
		/*
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		
		glBindTexture(GL_TEXTURE_2D,0);
	
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		
        glEnable(GL_TEXTURE_2D);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        */
        
		char buf[100];
		sprintf(buf,"world: %5.1f, %5.1f  window: %5.1f, %5.1f",
				world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
		DrawString(50,50,buf);

		const int bodies = (int) simulator.m_bodies.size();
		const int forces = (int) simulator.m_binary_forces.size();
		const int contacts = (int) simulator.m_contacts.size();
		const int other = (int) (simulator.m_global_forces.size() + simulator.m_unary_forces.size());
	    const int nobjs = (int) objs.size();
	    const int nstars = (int) starfield.size();
	    const int nstarflares = (int) starflares.size();
	    const int nsprites = (int) sprites.size();
	    
		sprintf(buf,"bodies = %d forces = %d contacts = %d other = %d objs = %d",
				bodies,forces,contacts,other,nobjs);
		DrawString(50,20,buf);
		OutputDebugString(buf);
		
	    /*
		Vec2 LeftStick = PS2Controller.left_stick(0);
		Vec2 RightStick = PS2Controller.right_stick(0);
		sprintf(buf,"left: %5.1f, %5.1f  right: %5.1f, %5.1f",
				LeftStick.x(),LeftStick.y(),RightStick.x(),RightStick.y());
		DrawString(50,20,buf);
		*/

		char *ButtonStrings[] =
		{
			"BUTTON_RIGHT_D",
			"BUTTON_RIGHT_U",
			"BUTTON_RIGHT_L",
			"BUTTON_RIGHT_R",
			"BUTTON_SHOULDER_R1",
			"BUTTON_SHOULDER_R2",
			"BUTTON_SHOULDER_L1",
			"BUTTON_SHOULDER_L2",
			"BUTTON_RSTICK",
			"BUTTON_LSTICK",
			"BUTTON_START",
			"BUTTON_SELECT",
		};

		int Y = 70;
		for(int ButtonIndex = 0;
			ButtonIndex < Joystick::NUM_BUTTONS;
			++ButtonIndex)
		{
			if(PS2Controller.button_is_down(ButtonIndex))
			{
				DrawString(50.f,(float)Y,ButtonStrings[ButtonIndex]);
				Y += 10;
			}
		}

		glDisable(GL_TEXTURE_2D);
		
		draw_obj_attribs(simulator,window_mouse,world_mouse);
		DrawTimingBars();
    }
}


/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
			if ( ! autozoom )
			{
				float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
				CameraZ *= 1.0f + d/10.0f;
				world_mouse = WorldFromWindowCoords(window_mouse);
			}
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) 
            {
            
				case ' ':
					paused = !paused;
					break;
					
				#ifdef DEBUG_MODE
				case VK_F6:
					LoadControls();
					break;
				case VK_F7:
					SaveControls();
					break;
				
				case VK_F4: {
					Reset();
                } break;     

                case VK_F2: {
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
                
                case VK_F3: {
					autozoom = !autozoom;
				} break;
            
				case VK_F1:
				{
					//debugstuff = !debugstuff;
					
					if ( GlobalsEditor == NULL )
					{
						GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 0, 400, 1000);
						EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
					}
					break;
				}
				#endif // DEBUG_MODE
            
                case VK_ESCAPE: {
					if ( MenuActive )
	                {
	                    AppQuit = true;
	                }
	                else
	                {
						MenuActive = true;
						GameOver = GameOverKillCounted = false;
						if ( players[0].ship )
							BlowUpObject(players[0].ship);
						if ( players[1].ship )
							BlowUpObject(players[1].ship);
	                }
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            Vec2 new_world_mouse = WorldFromWindowCoords(new_window_mouse);
			delta_window_mouse = new_window_mouse - window_mouse;
            delta_world_mouse = new_world_mouse - world_mouse;
			if(MovingCamera && ! autozoom)
			{
				CameraX -= delta_world_mouse.x();
				CameraY -= delta_world_mouse.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = new_world_mouse;
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

#ifdef DEBUG_MODE
    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
#else
    DWORD Style = WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
#endif

    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,NULL,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

   GetFontBase();
   
    // okay, ready to go
   
    Initialize();    // call the app
   
#ifdef DEBUG_MODE
    ShowWindow(WindowHandle,SW_SHOWDEFAULT); //|SW_SHOWDEFAULT);
#else
    ShowWindow(WindowHandle,SW_SHOWMAXIMIZED);
#endif

    BringWindowToTop(WindowHandle);
   
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}

/*
float ButtonScalar(bool b)
{
	if ( b )
		return 1.f;
	else
		return 0.f;
}
*/

float StickAngleScalar(Vec2 v,float offset,float last)
{
	if ( v.x() == 0.f && v.y() == 0.f )
		return last;

	float a = atan2f(-v.x(),v.y()) + offset;
	
	while ( a < (last - PI) )
		a += 2*PI;
	while ( a > (last + PI) )
		a -= 2*PI;
	
	return a;
}

bool NewKeyDown(const int vk)
{
	if (GetAsyncKeyState(vk) & 0x1)
		return true;

	return false;
}

bool IsKeyDown(const int vk)
{
	if (GetAsyncKeyState(vk) & 0x8000)
		return true;

	return false;
}

/*
float StickScalar(float x,cbControlOrder order)
{
	return fclamp(x,-1.f,1.f);
}

float MouseScalar(float v,float d,cbControlOrder order)
{
	switch(order)
	{
	case eControlValue:
		// -1 to 1 to [0,1]
		// @@ scale
		return v;
	case eControlDelta:
		// -1 to 1 is good
		// @@ scale
		return d;
	}	
	return 0.f;
}

float cbInput(const cbControl & control)
{
	const cbControlOptions & options = control.m_options;
	
	Vec2 LeftStick = PS2Controller.left_stick(options.m_gamepad);
	Vec2 RightStick = PS2Controller.right_stick(options.m_gamepad);
	
	switch(options.m_input)
	{
	case eLeftStickV:
		return StickScalar(LeftStick.y(),options.m_order);
		break;
	case eLeftStickH:
		return StickScalar(LeftStick.x(),options.m_order);
		break;
	case eRightStickV:
		return StickScalar(RightStick.y(),options.m_order);
		break;
	case eRightStickH:
		return StickScalar(RightStick.x(),options.m_order);
		break;
	case eLeftStickAngle:
		return StickAngleScalar(LeftStick,options.m_controlRest,control.m_value);
		break;
	case eRightStickAngle:
		return StickAngleScalar(RightStick,options.m_controlRest,control.m_value);
		break;
	case eButton_X:
		return ButtonScalar( PS2Controller.button_is_down(0,options.m_gamepad) );
		break;
	case eButton_O:
		return ButtonScalar( PS2Controller.button_is_down(1,options.m_gamepad) );
		break;
	case eButton_Square:
		return ButtonScalar( PS2Controller.button_is_down(2,options.m_gamepad) );
		break;
	case eButton_Triangle:
		return ButtonScalar( PS2Controller.button_is_down(3,options.m_gamepad) );
		break;
	case eMouseV:
		return MouseScalar(window_mouse.y(),delta_window_mouse.y(), options.m_order);
		break;
	case eMouseH:
		return MouseScalar(window_mouse.x(),delta_window_mouse.x(), options.m_order);
		break;
	case eKey:
	{
		char key = options.m_key[0];
		return ButtonScalar( IsKeyDown(key) );
		break;
	}
	case eNull:
		break;
	};
	
	return 0.f;
}

void cbControl::Tick()
{
	// find the Binary_force

	if ( m_options.m_name.empty() )
		return; // null

	Binary_Force * force = NULL;
	for(int i=0;i<forces.size();i++)
	{
		if ( forces[i]->name() == m_options.m_name )
		{
			force = forces[i];
			break;
		}
	}
	if ( force == NULL )
	{
		cbLog("couldn't hook up to force : %s\n",m_options.m_name.c_str());
		return;
	}
	
	float input = cbInput(*this);

	float value;
	
 	switch(m_options.m_order)
	{
		case eControlValue:
		{
			if ( m_options.m_input == eLeftStickAngle || m_options.m_input == eRightStickAngle )
			{
				value = input;
			}
			else
			{
				if ( input < 0.f )
					value = flerp(m_options.m_controlRest,m_options.m_controlLo,-input);
				else
					value = flerp(m_options.m_controlRest,m_options.m_controlHi,input);
			}
			break;	
		}
		case eControlDelta:
		{
			value = m_value + m_options.m_controlDeltaSpeed * input;
			value = fclamp(value,m_options.m_controlDeltaLo,m_options.m_controlDeltaHi);
			break;
		}
	}
	
	m_value = value;
	
	Spring_Force * sf = dynamic_cast<Spring_Force *>(force);
	Angle_Spring_Force * asf = dynamic_cast<Angle_Spring_Force *>(force);
	//Pt2Pt_Constraint * p2p = dynamic_cast<Pt2Pt_Constraint *>(force);
	
	//cbLog(" Value : %s\n",value );
	
	switch(m_options.m_type)
	{
	case eSpringLength:
		if ( sf )
			sf->m_rest_len = value;
		else
			cbLog("Error : Not a Spring : %s\n",m_options.m_name.c_str());
		break;
	case eAngle:
		if ( asf )
			asf->m_rest_len = value;
		else
			cbLog("Error : Not an Angle Spring : %s\n",m_options.m_name.c_str());
		break;
	
	}
	
}
*/

void TickControls()
{
/*
	for(int i=0;i<controls.size();i++)
	{
		controls[i].Tick();
	}
	*/
}

char * Filename = "controls.prop";

void SaveControls()
{
	vector<property *> temps;

	property * Parent = CreateProperty(SubGroupProp,"Parent");

	{
		property * prop = CloneProperty(GlobalProperties);
		SetPropertyGroup(prop,Parent);
	}

	for(int i=0;i<4;i++)
	{
		property * prop = CloneProperty(PROP_LOCAL_STRUCT(Keymap));
		ConvertStructToProperty(PROP_LOCAL_STRUCT(Keymap), 
								keymaps[i],
								prop);
                            
		SetPropertyGroup(prop,Parent);
		
		temps.push_back(prop);
	}

	WritePropertyToFile(Filename,Parent);
	
	// @@ ??
	DeleteProperty( Parent );
	// bad leak, but who cares
}

void LoadControls()
{
	property * Read = ReadPropertyFromFile(Filename);
	
	property * First = FirstSubProperty(Read);
	
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            First,
                            &Globals);

	UpdateControlCounts();

	property * Next = First;
	
	for(int i=0;i<4;i++)
	{
		Next = NextProperty(Next);
		if ( Next == NULL )
			break;

		ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "Keymap"),
                            Next,
                            keymaps[i]);
		
	}
	
	RefreshEditors();
}

void SpawnShrapnel(game_object * from,const int shrapnel_count,float ShrapnelOffset,
							float ShrapnelMass,float ShrapnelVelocityFraction,
							float ShrapnelVelocity,GLuint tex,
							float Scale)
{
	Vec2 pos = from->rb->x;
	Vec2 base_v = from->rb->v;
		
	for(int i=0;i<shrapnel_count;i++)
	{		
		float a = i * 2 * PI / shrapnel_count;
		Vec2 facing = Vec2( cosf(a), sinf(a) );
		Vec2 poop_pos = pos + facing * ShrapnelOffset;
		
		Rigid_Body * poop = new Rigid_Body(
			make_Body_Box_as_Box(Vec2(-0.5f, -0.5f), Vec2(0.5f, 0.5f)),
			poop_pos,
			0.0f,
			1.0f/ShrapnelMass,
			10.0f
		);
		
		poop->v = base_v * ShrapnelVelocityFraction + ShrapnelVelocity * facing;
	
		poop->update_derived_state();
	
		objs.push_back( new game_object(NULL,poop) );
		//memcpy(objs.back()->color,color,3*sizeof(float));
		objs.back()->texture = tex;
		objs.back()->death_time = objs.back()->spawn_time + Globals.PoopLifeTime;
		objs.back()->is_shrapnel = true;
		objs.back()->quad_size *= Scale;
	}
}
	
void                   
ControlShipSub(Player * player,int pad,
	bool & is_pooping)
{
	if ( ! player->ship )
		return;
	Rigid_Body * ship = player->ship->rb;
	if ( ! ship )
		return;

	player->ship->is_scooping_poops = false;

	double now = GetTime();
	
	Vec2 pos = ship->x;
	Vec2 facing = get_facing(ship);
		
	if ( (now - player->started_starfighter_time) < Globals.StarFighterDuration )
	{
		ship->O += Globals.StarFighterAngleRate;
		ship->w = 0.f;
		
		zsoundPlay(snd_starfighterMissile);
			
		game_object * missile = MakeMissile();
		objs.push_back(missile);
		
		missile->color[0] = 1;
		missile->color[1] = 0;
		missile->color[2] = 0;
		
		missile->is_armed_missile = true;
		
		Vec2 impulse = facing * Globals.ShootImpulse;

		missile->rb->x = pos + facing * Globals.ShootOffset;
		missile->rb->O = ship->O;

		missile->rb->v = ship->v;

		missile->rb->v += missile->rb->inv_m * impulse;
	
		missile->rb->update_derived_state();
			
		missile->death_time = missile->spawn_time + Globals.StarFighterMissileLifeTime;
		
		missile->texture = tex_missile1;
			
		missile->quad_size *= Globals.SpriteEmpScale;
			
		player->last_disable_time = now;
		
		ship->update_derived_state();
		return;
	}
		
	if ( (now - player->last_disable_time) < Globals.MissileDisableDuration )
		return;

	player->energy += Globals.EnergyRegenRate;

	//Vec2 move = PS2Controller.left_stick(pad);
	//Vec2 face = PS2Controller.right_stick(pad);
	Vec2 lstick = PS2Controller.left_stick(pad);
	Vec2 rstick = PS2Controller.right_stick(pad);

	// @@ more speed when move & face align

	//Vec2 impulse = move * Globals.ThrustImpulse;

	//obj->rb->v += obj->rb->inv_m * impulse;

	if ( lstick.len() > 0.f )
	{
		float oldAngle = ship->O.z();
		float newAngle = StickAngleScalar(lstick,0.f,oldAngle);
		float angle = LerpedDrive(oldAngle,newAngle, Globals.TurnRate, 1.f);
		ship->O = angle;	
		ship->w = 0.f;
	}
	
	float keyTurnRate = 0.1f;
	if ( IsKeyDown(keys[pad].turnLeft) )
	{
		ship->O += keyTurnRate;	
		ship->w = 0.f;
	}
	if ( IsKeyDown(keys[pad].turnRight) )
	{
		ship->O -= keyTurnRate;	
		ship->w = 0.f;
	}
	
	facing = get_facing(ship);
	
	// @@ should lerp the grapple gun facing
	if ( rstick.len() > 0.f )
	{
		player->grapple_gun_facing = rstick;
	}
	else
	{
		player->grapple_gun_facing = facing;
	}
				
	is_pooping = PS2Controller.button_is_down(buttons[pad].thrust,pad) || IsKeyDown(keys[pad].thrust);
	if ( is_pooping &&
		(now > player->last_thrust_time + Globals.ThrustInterval) )
	{
		player->last_thrust_time = now;
 
		Vec2 poop_pos = pos + facing * (- Globals.PoopOffset);
		
		zsoundPlay(snd_poop);
		
		Rigid_Body * poop = new Rigid_Body(
			make_Body_Box_as_Box(Vec2(-0.5f, -0.5f), Vec2(0.5f, 0.5f)),
			poop_pos,
			0.0f,
			1.0f/Globals.PoopMass,
			//10.0f
			0.2f // inverse moment
		);
		
		Vec2 impulse = facing * Globals.ThrustImpulse;

		//poop->v = ship->v;
		
		ship->v += ship->inv_m * impulse;
		poop->v -= poop->inv_m * impulse;
	
		poop->update_derived_state();
		
		objs.push_back( new game_object(NULL,poop) );
		objs.back()->is_poop = true;
		objs.back()->texture = tex_poop;
		objs.back()->color[0] = player->ship->color[0];
		objs.back()->color[1] = player->ship->color[1];
		objs.back()->color[2] = player->ship->color[2];
		objs.back()->death_time = objs.back()->spawn_time + Globals.PoopLifeTime;
	}
	
	bool down1 = PS2Controller.button_is_down(buttons[pad].missile1,pad) || IsKeyDown(keys[pad].missile1);	
	bool down2 = PS2Controller.button_is_down(buttons[pad].missile2,pad) || IsKeyDown(keys[pad].missile2);	
	bool ram_scoop = PS2Controller.button_is_down(buttons[pad].ramscoop,pad) || IsKeyDown(keys[pad].ramscoop);
	bool grapple = PS2Controller.button_is_down(buttons[pad].grapple,pad) || IsKeyDown(keys[pad].grapple);	
	
	bool starfighter = down1 && down2 && ram_scoop && grapple;
		
	if ( starfighter )
	{
		down1 = down2 = ram_scoop = false;
	}
	
	bool fire2 = (down2 && ! player->shrapnelly);
	bool blow2 = (!down2 && player->shrapnelly);
	
	
	if ( ! grapple )
	{
		// release the grapple head
		if ( player->grapple_force )
		{
			DeleteForce(player->grapple_force);
		}
		if ( player->grapple_head )
		{
			DeleteObject(player->grapple_head);
		}
		player->grapple_head = NULL;
		player->grapple_force = NULL;
	}
	if ( grapple && player->grapple_head == NULL && player->grapple_force == NULL )
	{
		// fire the grapple header

		Vec2 poop_pos = pos + player->grapple_gun_facing * Globals.ShootOffset;
		
		zsoundPlay(snd_grappleFire);
			
		game_object * missile = MakeMissile();
		objs.push_back(missile);
		
		missile->color[0] = 0;
		missile->color[1] = 1;
		missile->color[2] = 0;
			
		missile->is_grapple_head = true;
		
		Vec2 impulse = player->grapple_gun_facing * Globals.ShootImpulse;

		missile->rb->x = poop_pos;
		missile->rb->O = ship->O;
		missile->rb->v = ship->v;
		missile->rb->v += missile->rb->inv_m * impulse;
	
		missile->rb->update_derived_state();
		
		player->grapple_head = missile;
	}
		
	if ( ram_scoop && ! player->shrapnelly )
	{
		player->ship->ram_scooping = true;
		// do scooping
		// look for poops
		for(int i=(int)objs.size()-1;i>=0;i--)
		{
			if ( ! objs[i]->is_poop )
				continue;
			Vec2 poop_pos = objs[i]->rb->x;
			Vec2 d = poop_pos - pos;
			float l2 = d.len2();
			if ( l2 < fsquare(Globals.RamScoopRadius) )
			{
				float dot = d.norm() * facing;
				if ( dot >= cosf(Globals.RamScoopAngle) )
				{
					zsoundPlay(snd_scoopPoop,0,0.5f);
		
					// eat it !
					player->energy += Globals.EnergyBoostRamScoop;

					player->ship->is_scooping_poops = true;
		
					ship->v += facing * Globals.RamScoopVelocity;

					delete objs[i];
					objs[i] = objs.back();
					objs.pop_back();
				}
			}
		}
	}
	else
	{
		player->ship->ram_scooping = false;
	}
		
	
	if ( (down1 || fire2) &&
		(now > player->last_shoot_time + Globals.ShootInterval) )
	{
		player->last_shoot_time = now;
		
		Vec2 poop_pos = pos + facing * Globals.ShootOffset;
					
		game_object * missile = MakeMissile();
		objs.push_back(missile);
		
		if ( down1 )
		{
			zsoundPlay(snd_fire1);
		
			missile->texture = tex_missile1;
			
			missile->quad_size *= Globals.SpriteEmpScale;
			
			missile->is_armed_missile = true;
		}
		else
		{
			zsoundPlay(snd_fire2);

			missile->texture = tex_missile2;

			missile->quad_size *= Globals.SpriteMissileScale;
			
			player->shrapnelly = missile;
		}
		
		Vec2 impulse = facing * Globals.ShootImpulse;

		missile->rb->x = poop_pos;
		missile->rb->O = ship->O;

		missile->rb->v = ship->v;

		ship->v -= ship->inv_m * impulse;
		missile->rb->v += missile->rb->inv_m * impulse;
	
		missile->rb->update_derived_state();
			
		missile->death_time = missile->spawn_time + Globals.MissileLifeTime;
	}
	
	if ( blow2 && player->shrapnelly )
	{
		zsoundPlay(snd_missileShrapnel);
		
		const int shrapnel_count = 10;
		float color[3] = { 1,1,1 };
		SpawnShrapnel(player->shrapnelly,shrapnel_count,Globals.ShrapnelOffset,
								Globals.ShrapnelMass,Globals.ShrapnelVelocityFraction,
								Globals.ShrapnelVelocity,tex_shrapnel2,Globals.SpriteMissileShrapnelScale);
		
		ASSERT( player->shrapnelly != NULL );
		DeleteObject(player->shrapnelly);
		player->shrapnelly = NULL;	
	}
	
	if ( starfighter && (player->energy/Globals.EnergyMax) >= 0.99f )
	{
		player->energy = 0.;
		player->started_starfighter_time = now;
		player->last_disable_time = now;
	}
				
	player->energy = min(player->energy,Globals.EnergyMax);
	ship->update_derived_state();

}

void DoLoop(int & zs,bool desired,char * snd)
{
	bool cur = zs && zsoundIsPlaying(zs);
	if ( desired && ! cur )
	{
		zs = zsoundPlay(snd,1);
	}
	else if ( ! desired && cur )
	{
		zsoundStop(zs);
		zs = 0;
	}
}

void ControlShip(Player * player,int pad)
{
	bool is_pooping = false;
	ControlShipSub(player,pad,is_pooping);
	player->is_pooping = is_pooping;
	
	bool is_scooping = player->ship->ram_scooping;
	bool is_grappling = player->grapple_force != NULL;

	static int zs_poop = zsoundPlay(snd_loop_poop,1);
	static int zs_scoop = 0;
	static int zs_grapple = 0;

	//DoLoop(zs_poop,is_pooping,snd_loop_poop);
	DoLoop(zs_scoop,is_scooping,snd_loop_scoop);
	DoLoop(zs_grapple,is_grappling,snd_loop_grapple);
	
	float cur_volume;
	int l; float p,f;
	zsoundGetStats(zs_poop,l,cur_volume,p,f);
	float desired_volume = is_pooping ? 1.f : 0.f;

	float new_volume = LerpedDrive(cur_volume,desired_volume,1.f,dt);
	zsoundChangeVolume(zs_poop,new_volume);
}

float funitrand()
{
	return ((float)rand()/RAND_MAX);
}

float frandpos()
{
	return (funitrand() - 0.5f) * 2.f * Globals.StarSpawnRadius;
}

void SpawnStar()
{
	starfield.push_back( Star() );
	starfield.back().v = Vec2(0,0);
	starfield.back().x = Vec2( frandpos(), frandpos() );
}

bool TickStar(Star & star)
{

	// apply gravity
	Vec2 to = (SunPos - star.x);
	float len2 = to.len2();
	Vec2 accel = to * (Globals.SunGravity/len2); //*len));

	star.v += accel;
	star.x += star.v * dt;
	
	// check for hitting the sun
	return ( len2 < SunRadius*SunRadius );
}

void BlowUpObject(game_object * obj)
{
	if ( obj == NULL )
		return;
	if ( obj == players[0].ship )
	{
		players[0].ship = NULL;
	}
	else
	{
		players[1].ship = NULL;
	}

	//SpawnShrapnel(obj,30,2.f,100.f,0.5f,20.f,obj->color);

	SpawnExplosion(obj->rb->x,tex_explode,4.f,4.f);
	
	zsoundPlay(snd_shipExplosion);
	
	obj->color[0] = 0;
	obj->color[1] = 0;
	obj->color[2] = 0;
	
	//DeleteObject(obj);
}
		
bool TickObject(game_object * obj)
{
	if ( obj->rb->inv_m == 0.f )
		return false;
		
	// apply gravity
	Vec2 pos = obj->rb->x;
	Vec2 to = (SunPos - pos);
	float len = to.len();
	Vec2 accel = to * (1.f/(len*len)); //*len));
	
	obj->rb->v += accel * Globals.SunGravity;
	
	// check for hitting the sun
	if ( len < SunRadius )
	{
		// kill me !
		// @@ spawn cool explosion
		
		if ( IsCircle(obj->rb) )
		{
			// planet; 
			return false;
		}
		
		if ( obj == players[0].ship || obj == players[1].ship )
		{
			// @@ a player died !  Reset
			BlowUpObject(obj);

			if ( ! GameOver )
			{
				zsoundPlay(snd_endtro);
				GameOver = true;
				GameOverKillCounted = false;
				last_gameover_time = now;
			}
			return true;
		}
		
		return true;
	}
	
	return false;
}

Vec2 rand_unit_box()
{
	float angle = funitrand() * 2 * PI;
	Vec2 n( cosf(angle), sinf(angle) );
	// project n to a unit square :
	float ax = fabsf(n.x());
	float ay = fabsf(n.y());
	if ( ax > ay )
	{
		n = Vec2( fsign(n.x()), n.y()/ax );
	}
	else
	{
		n = Vec2( n.x()/ay, fsign(n.y()) );
	}
	return n;
}

void SpawnPlanet()
{
	Vec2 p1 = players[0].ship ? players[0].ship->rb->x : SunPos;
	Vec2 p2 = players[1].ship ? players[1].ship->rb->x : SunPos;
	
	// 
	for(;;)
	{
		Vec2 n = rand_unit_box();
		Vec2 pos = SunPos + n * Globals.PlanetSpawnDist;
		
		// if p is too close to a player, try again
		float dist1 = (pos - p1).len();
		float dist2 = (pos - p2).len();
		if ( dist1 < Globals.PlanetSpawnRadius * 4 ||
			dist2 < Globals.PlanetSpawnRadius * 4 )
		{
			continue;
		}
		
		// make it :

		Vec2 p( n.y(), - n.x() );				
		p = p.norm();
		p *= (funitrand() - 0.5f) * 2.f * Globals.PlanetVelocity;
				
		float r = (funitrand() + 1.f) * 0.5f * Globals.PlanetSpawnRadius;
				
		float mass = r*r* 1.f;		
				
		Rigid_Body * poop = new Rigid_Body(
			make_Body_Sphere(r),
			pos,
			0.0f,
			1.0f/mass,
			0.0f // inv I
		);
		
		poop->v = p;
		// planets don't collide with the edge of the universe :
		poop->m_Pbody->m_collision_plane = 2;
		
		poop->update_derived_state();
		
		objs.push_back( new game_object(NULL,poop) );
		objs.back()->color[0] = 0.f;
		objs.back()->color[1] = 1.f;
		objs.back()->color[2] = 0.f;	
		
		objs.back()->rb->w.set_z( funitrand()*3/20 );
		
		objs.back()->texture = tex_planet->GetRandom();  

		break;
	}
}

void SpawnExplosion(Vec2 contactPos,TextureSet * tex,float scale,float duration)
{
	sprites.push_back( Sprite() );
	Sprite & s = sprites.back();
	s.angle = funitrand() * 2*PI;
	s.pos = contactPos;
	s.duration = duration;
	s.numLoops = 1;
	s.start = now;
	s.tex = tex;
	s.w = s.h = 20.f * scale;
}
